/* run -- Wrapper program for console mode programs under Windows(TM)
 * Copyright (C) 1998  Charles S. Wilson
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

/*
 * This program is based on the runemacs.c distributed with XEmacs 21.0
 *
 * Simple program to start gnu-win32 X11 programs (and native XEmacs) 
 * with its console window hidden.
 *
 * This program is provided purely for convenience, since most users will
 * use XEmacs in windowing (GUI) mode, and will not want to have an extra
 * console window lying around. Ditto for desktop shortcuts to gnu-win32 
 * X11 executables.
 */
#if HAVE_CONFIG_H
#  include "config.h"
#endif

#ifndef WIN32
#define WIN32
#endif

/* pull in WinXP function declarations, especially AttachConsole */
#define _WIN32_WINNT 0x0501

#include <windows.h>
#include <winuser.h>
#include <string.h>
#include <malloc.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>

#include "run.h"

#if defined(__CYGWIN__)
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <sys/cygwin.h>
 #include <sys/unistd.h>
#if 0
WinMainCRTStartup() { mainCRTStartup(); }
#endif
#else
 #include <direct.h>
#endif

DWORD os_version;
char buffer[1024];

int WINAPI
WinMain (HINSTANCE hSelf, HINSTANCE hPrev, LPSTR cmdline, int nShow)
{
   int wait_for_child = FALSE;
   int compact_invocation = FALSE;
   DWORD ret_code = 0;


   char execname[FILENAME_MAX];
   char execpath[MAX_PATH];
   char* argv[MAX_ARGS+1]; /* leave extra slot for compact_invocation argv[0] */
   int argc;
   int i,j;
   char exec[MAX_PATH + FILENAME_MAX + 100];
   char cmdline2[MAX_ARGS * MAX_PATH];
   DWORD vers = GetVersion ();
   os_version = (LOBYTE (LOWORD (vers)) << 8) | HIBYTE (LOWORD (vers));

   cmdline = GetCommandLine();
   /* strip program name. Maybe quoted? */
   if (*cmdline == '\"')
   {
       char *s = cmdline + 1;
       /* search next quote character. 
        * don't care if " is a literal in the name */
       while (*s != 0 && *s != '\"')
           s++;
       if (*s == '\"') {
           char* token = cmdline + 1;
           /* we have found the starting quote (*cmdline) and the 
            * ending quote (*s). everything between is now the first
            * argument */
           s++;

           /* find the start of the next argument */
           while (*s != 0 && strchr(" \t\n", *s) != NULL)
           {
               s++;
           }

           /* continue processing with the remaining cmdline */
           cmdline = s;
       }
   } else {
       char *s = cmdline;
       /* find the end of this argument */
       while (*s != 0 && strchr(" \t\n", *s) == NULL)
           s++;
       while (*s != 0 && strchr(" \t\n", *s) != NULL)
           s++;
       cmdline = s;
   }

   compact_invocation = get_exec_name_and_path(execname,execpath);

   if (compact_invocation)
   {
      argv[0] = execname;
      argc = parse_cmdline_to_arg_array(&(argv[1]),cmdline);
      argc++;
   }
   else
   {
      argc = parse_cmdline_to_arg_array(argv,cmdline);
      if (argc >= 1)
      {
         if (argv[0] && *argv[0] == '"')
         {
           int l;
           strncpy(execname,argv[0]+1,FILENAME_MAX);
           execname[FILENAME_MAX-1] = 0;
           l = strlen(execname);
           if (l > 0 && execname[l-1] == '"')
             execname[l-1] = 0;
         } else
         {
           strncpy(execname,argv[0],FILENAME_MAX);
           execname[FILENAME_MAX-1] = 0;
         }
      }
   }
   /* at this point, execpath is defined, as are argv[] and execname */
#ifdef DEBUG
   j = sprintf(buffer,"\nexecname : %s\nexecpath : %s\n",execname,execpath);
   for (i = 0; i < argc; i++)
      j += sprintf(buffer+j,"argv[%d]\t: %s\n",i,argv[i]);
   Trace((buffer));
#endif

   if (execname == NULL)
      error("you must supply a program name to run");

#if defined(__CYGWIN__)
   /* this insures that we search for symlinks before .exe's */
   if (compact_invocation)
      strip_exe(execname);
#endif

   process_execname(exec,execname,execpath);
   Trace(("exec\t%s\nexecname\t%s\nexecpath\t%s\n",
         exec,execname,execpath));

   wait_for_child = build_cmdline(cmdline2,exec,argc,argv);
   Trace((cmdline2));

   xemacs_special(exec);
   ret_code = start_child(cmdline2,wait_for_child);
   if (compact_invocation)
      for (i = 1; i < argc; i++) /* argv[0] was not malloc'ed */
         free(argv[i]);
   else
      for (i = 0; i < argc; i++)
         free(argv[i]);
   return (int) ret_code;
}

/* Copy cygwin environment variables to the Windows environment if they're not
 * already there. */
static void setup_win_environ(void)
{
    char **envp = environ;
    char *var, *val;
    char curval[2];

    while (envp && *envp) {
        var = strdup(*envp++);
        val = strchr(var, '=');
		if (val != NULL)
		{
			*val++ = '\0';

			if (GetEnvironmentVariable(var, curval, 2) == 0 &&
			    GetLastError() == ERROR_ENVVAR_NOT_FOUND)
			{
				SetEnvironmentVariable(var, val);
			}
		}
        free(var);
    }
}
BOOL setup_invisible_console()
{
   HWINSTA h, horig;
   USEROBJECTFLAGS oi;
   DWORD len;
   BOOL b = FALSE;
   HMODULE lib = NULL;
   HWINSTA WINAPI (*GetProcessWindowStationFP)(void) = NULL;
   BOOL WINAPI (*GetUserObjectInformationFP)(HANDLE, int, PVOID, DWORD, PDWORD) = NULL;
   HWINSTA WINAPI (*CreateWindowStationFP)(LPCWSTR, DWORD, DWORD, LPSECURITY_ATTRIBUTES) = NULL;
   BOOL WINAPI (*SetProcessWindowStationFP)(HWINSTA) = NULL;
   BOOL WINAPI (*CloseWindowStationFP)(HWINSTA) = NULL;

   /* until we have a mechanism of determining whether a given HANDLE 
    * returned by GetStdHandles actually derives from a console, 
    * unconditionally call FreeConsole() on all OSes under all conditions.
    * See comments in configure_startupinfo(). 
    */
   FreeConsole();

   /* First, set up function pointers */
   if (lib = LoadLibrary ("user32.dll"))
   {
       GetProcessWindowStationFP = (HWINSTA WINAPI (*)(void))
           GetProcAddress (lib, "GetProcessWindowStation");
       GetUserObjectInformationFP = (BOOL WINAPI (*)(HANDLE, int, PVOID, DWORD, PDWORD))
           GetProcAddress (lib, "GetUserObjectInformationW"); /* ugly! */
       CreateWindowStationFP = (HWINSTA WINAPI (*)(LPCWSTR, DWORD, DWORD, LPSECURITY_ATTRIBUTES))
           GetProcAddress (lib, "CreateWindowStationW"); /* ugly */
       SetProcessWindowStationFP = (BOOL WINAPI (*)(HWINSTA))
           GetProcAddress (lib, "SetProcessWindowStation");
       CloseWindowStationFP = (BOOL WINAPI (*)(HWINSTA))
           GetProcAddress (lib, "CloseWindowStation");

       if (GetProcessWindowStationFP &&
            GetUserObjectInformationFP &&
            CreateWindowStationFP &&
            SetProcessWindowStationFP &&
            CloseWindowStationFP)
       {
           /* Then, do the work */
           FreeConsole();
           h = horig = (*GetProcessWindowStationFP)();
           if (!horig
               || !(*GetUserObjectInformationFP) (horig, UOI_FLAGS, &oi, sizeof (oi), &len )
               || !(oi.dwFlags & WSF_VISIBLE))
           {
               b = AllocConsole();
           }
           else
           {
               h = (*CreateWindowStationFP) (NULL, 0, STANDARD_RIGHTS_READ, NULL);
               if (h)
               {
                   b = (*SetProcessWindowStationFP) (h);
               }
               b = AllocConsole();
               if (horig && h && h != horig && (*SetProcessWindowStationFP) (horig))
               {
                   (*CloseWindowStationFP) (h);
               }
           }
           return b;
       }
   }
   /* otherwise, fail */
   return FALSE;
}

/* returns FALSE only on error conditions (not impl) */
BOOL configure_startupinfo(STARTUPINFO* psi, BOOL bHaveInvisConsole,
                           BOOL *bUsingPipes,
                           HANDLE* hpToChild,  HANDLE* hpFromChild,
                           HANDLE* hpToParent, HANDLE* hpFromParent)
{
    CONSOLE_CURSOR_INFO cci;
    SECURITY_ATTRIBUTES handle_attrs;
    HANDLE hpStdInput[2];
    HANDLE hpStdOutput[2];

    ZeroMemory (psi, sizeof (STARTUPINFO));
    psi->cb = sizeof (STARTUPINFO);
    psi->hStdInput   = GetStdHandle(STD_INPUT_HANDLE);
    psi->hStdOutput  = GetStdHandle(STD_OUTPUT_HANDLE);
    psi->hStdError   = GetStdHandle(STD_ERROR_HANDLE);
    psi->dwFlags     = STARTF_USESHOWWINDOW | STARTF_USESTDHANDLES;
    psi->wShowWindow = SW_HIDE;

    /* foo() is some magic mechanism for determining that the HANDLEs 
     * returned by GetStdHandle() are from a console, and not redirected
     * or ptys of some sort.  If we have such a mechanism, then the 
     * unconditional FreeConsole() at the top of setup_invisible_console()
     * should be removed.
     */
/*
    if (foo())
    {
       *bUsingPipes = FALSE;
       return TRUE;
    }
*/

    /* but for now, the only way we KNOW we have a console is
     * if we created it ourselves
     */
    if (bHaveInvisConsole)
    {
       *bUsingPipes = FALSE;
       return TRUE;
    }

    /* otherwise, set up pipes */
    *bUsingPipes = TRUE;

    handle_attrs.nLength = sizeof(SECURITY_ATTRIBUTES);
    handle_attrs.bInheritHandle = TRUE;
    handle_attrs.lpSecurityDescriptor = NULL;

    /* create a pipe for child's stdin.  Don't allow child to */
    /* inherit the write end of the pipe.                     */
    CreatePipe (&hpStdInput[0], &hpStdInput[1], &handle_attrs, 0);
    SetHandleInformation ( hpStdInput[1], HANDLE_FLAG_INHERIT, 0);

    /* create a pipe for child's stdout.  Don't allow child to */
    /* inherit the read end of the pipe.                       */
    CreatePipe (&hpStdOutput[0], &hpStdOutput[1], &handle_attrs, 0);
    SetHandleInformation ( hpStdOutput[0], HANDLE_FLAG_INHERIT, 0);

    psi->hStdInput   = hpStdInput[0];
    psi->hStdOutput  = hpStdOutput[1];
    psi->hStdError   = hpStdOutput[1];

    *hpToChild   = hpStdInput[1];
    *hpFromChild = hpStdOutput[0];
    *hpToParent  = hpStdOutput[1];
    *hpFromParent= hpStdInput[0];

    return TRUE;
}
int start_child(char* cmdline, int wait_for_child)
{
   STARTUPINFO start;
   PROCESS_INFORMATION child;
   DWORD retval;
   BOOL bFuncReturn;
   BOOL bHaveInvisConsole;
   BOOL bUsingPipes;
   HANDLE hToChild, hFromChild;
   HANDLE hToParent, hFromParent;
   BOOL WINAPI (*AttachConsoleFP)(DWORD) = NULL;
   HWND WINAPI (*GetConsoleWindowFP)(VOID) = NULL;

   setup_win_environ();

#if (defined (__CYGWIN__) && !HAVE_DECL_CYGWIN_CONV_PATH) || defined(__MINGW32__)
   /* mingw or cygwin-1.5: work around bug in Windows 7, but also
    * employ on XP and above. This means that setup_invisible_console()
    * is now used only on <= Win2k */
   if (os_version >= 0x0501)
     {
       HMODULE lib = GetModuleHandle ("kernel32.dll");
       AttachConsoleFP = (BOOL WINAPI (*)(DWORD))
           GetProcAddress (lib, "AttachConsole");
       GetConsoleWindowFP = (HWND WINAPI (*)(VOID))
           GetProcAddress (lib, "GetConsoleWindow");
       if (!AttachConsoleFP || !GetConsoleWindowFP)
           os_version = 0;
     }
#endif

#ifdef DEBUG_FORCE_PIPES
   bHaveInvisConsole = FALSE;
   FreeConsole();
#elif defined (__CYGWIN__) && HAVE_DECL_CYGWIN_CONV_PATH
   /* cygwin-1.7 */
   bHaveInvisConsole = TRUE;
#else
   /* mingw or cygwin-1.5: work around bug in Windows 7, but also
    * employ on XP and above. This means that setup_invisible_console()
    * is now used only on <= Win2k */
   bHaveInvisConsole = os_version >= 0x0501 ? TRUE : setup_invisible_console();
#endif

   if (!configure_startupinfo(&start, bHaveInvisConsole,
                              &bUsingPipes,
                              &hToChild, &hFromChild,
                              &hToParent, &hFromParent))
      error("could not start %s",cmdline);

#ifdef DEBUG
   message("Using Pipes: %s", (bUsingPipes ? "yes" : "no"));
#endif

   ZeroMemory( &child, sizeof(PROCESS_INFORMATION) );

   bFuncReturn = CreateProcess (NULL,
       cmdline, /* command line                        */
       NULL,    /* process security attributes         */
       NULL,    /* primary thread security attributes  */
       TRUE,    /* handles are inherited,              */
#if defined(__CYGWIN__) && HAVE_DECL_CYGWIN_CONV_PATH
       CREATE_NO_WINDOW,
#else
       0,       /* creation flags                      */
#endif
       NULL,    /* use parent's environment            */
       NULL,    /* use parent's current directory      */
       &start,  /* STARTUPINFO pointer                 */
       &child); /* receives PROCESS_INFORMATION        */

   if (bUsingPipes)
   {
       CloseHandle(hFromParent);
       CloseHandle(hToParent);
   }

   if (bFuncReturn)
   {
      if (wait_for_child)
      {
          if (bUsingPipes)
          {
              HANDLE handles[2] = { child.hProcess, hFromChild };
              COMMTIMEOUTS timeouts;

              /* only read bytes that are ready, and return immediately */
              GetCommTimeouts(hFromChild, &timeouts);
              timeouts.ReadIntervalTimeout = MAXDWORD;
              timeouts.ReadTotalTimeoutMultiplier = 0;
              timeouts.ReadTotalTimeoutConstant = 0;
              SetCommTimeouts(hFromChild, &timeouts);

              while (WaitForMultipleObjects (2, handles, FALSE, INFINITE) == WAIT_OBJECT_0 + 1)
              {
                  char buffer[1024];
                  DWORD dwRead;
                  ReadFile (hFromChild, buffer, 1024, &dwRead, NULL);
              }
          }
          else
          {
              WaitForSingleObject (child.hProcess, INFINITE);
          }
          GetExitCodeProcess (child.hProcess, &retval);
      }
#if (defined(__CYGWIN__) && !HAVE_DECL_CYGWIN_CONV_PATH) || defined(__MINGW32__)
     /* mingw or cygwin-1.5: work around bug in Windows 7, but also
      * employ on XP and above. This means that setup_invisible_console()
      * is now used only on <= Win2k */
      if (os_version >= 0x0501)
      {
        FreeConsole ();
        (*AttachConsoleFP) (child.dwProcessId);
        SetParent ((*GetConsoleWindowFP) (), HWND_MESSAGE);
      }
#endif
      CloseHandle (child.hThread);
      CloseHandle (child.hProcess);
      if (bUsingPipes)
      {
          CloseHandle (hFromChild);
          CloseHandle (hToChild);
      }
   }
   else
   {
      if (bUsingPipes)
      {
          CloseHandle (hFromChild);
          CloseHandle (hToChild);
      }
      error("could not start %s",cmdline);
   }
   return retval;
}
void xemacs_special(char* exec)
{
  /*
   * if we're trying to run xemacs, AND this file was in %emacs_dir%\bin,
   * then set emacs_dir environment variable 
   */
   char* p;
   char* p2;
   char exec2[MAX_PATH + FILENAME_MAX + 100];
   char tmp[MAX_PATH + FILENAME_MAX + 100];
   strcpy(exec2,exec);
   /* this depends on short-circuit evaluation */
   if ( ((p = strrchr(exec2,'\\')) && stricmp(p,"\\xemacs") == 0) ||
        ((p = strrchr(exec2,'/')) && stricmp(p,"/xemacs") == 0) ||
        ((p = strrchr(exec2,'\\')) && stricmp(p,"\\xemacs.exe") == 0) ||
        ((p = strrchr(exec2,'/')) && stricmp(p,"/xemacs.exe") == 0) )
   {
      if ( ((p2 = strrchr(p, '\\')) && stricmp(p2, "\\bin") == 0) ||
           ((p2 = strrchr(p, '/')) && stricmp(p2, "/bin") == 0) )
      {
         *p2 = '\0';
#if defined(__CYGWIN__)
         CYGWIN_CONV_TO_POSIX_PATH((exec2,tmp));
         strcpy(exec2,tmp);
#else /* NATIVE xemacs DOS-style paths with forward slashes */
         for (p = exec2; *p; p++)
            if (*p == '\\') *p = '/';
#endif
         SetEnvironmentVariable ("emacs_dir", exec2);
      }
   }
}
int build_cmdline(char* new_cmdline, char* exec, int argc, char* argv[])
{
   int retval = FALSE;
   int first_arg = 1;
   int i;
   int char_cnt = 0;
   /*
    * look for "-wait" as first true argument; we'll apply that ourselves
    */
   if ((argc >= 2) && (stricmp(argv[1],"-wait") == 0))
   {
      retval = TRUE;
      first_arg++;
   }

   char_cnt = strlen(exec);
   for (i = first_arg; i < argc; i++)
      char_cnt += strlen(argv[i]);
   if (char_cnt > MAX_ARGS*MAX_PATH) /* then we ran out of room */
      error("command line too long -\n%s",new_cmdline);

   strcpy(new_cmdline,exec);
   for (i = first_arg; i < argc; i++)
   {
      strcat(new_cmdline," ");
      strcat(new_cmdline,argv[i]);
   }
   return retval;
}
/* process exec_arg : if it
 * NATIVE:
 *  1) starts with '\\' or '/', it's a root-path and leave it alone
 *  2) starts with 'x:\\' or 'x:/', it's a root-path and leave it alone
 *  3) starts with '.\\' or './', two possible meanings:
 *       1) exec is in the current directory
 *       2) exec in same directory as this program
 *  4) otherwise, search path (and _prepend_ "." to the path!!!)
 *  5) convert all '/' to '\\'
 * CYGWIN
 *  1) starts with '\\' or '/', it's a root-path and leave it alone
 *  2) starts with 'x:\\' or 'x:/', it's a root-path and leave it alone
 *  3) starts with '.\\' or './', two possible meanings:
 *       1) exec is in the current directory
 *       2) exec in same directory as this program
 *  4) otherwise, search path (and _prepend_ "." to the path!!!)
 *  5) convert to cygwin-style path to resolve symlinks within the pathspec
 *  6) check filename: if it's a symlink, resolve it by peeking inside
 *  7) convert to win32-style path+filename since we're using Windows 
 *       createProcess() to launch
 */
void process_execname(char *exec, const char* execname,const char* execpath )
{
   char* orig_pathlist;
   char* pathlist;
   char exec_tmp[MAX_PATH + FILENAME_MAX + 100];
   char exec_tmp2[MAX_PATH + FILENAME_MAX + 100];
   char buf[MAX_PATH + FILENAME_MAX + 100];
   int i,j;

   int len = 0;
   /* 
    * STARTS WITH / or \ 
    * execpath NOT used
    */
   if ((execname[0] == '\\') || (execname[0] == '/'))
   {
#if defined(__CYGWIN__)
      strcpy(exec_tmp,execname);
#else
      exec_tmp[0] = ((char) (_getdrive() + ((int) 'A') - 1));
      exec_tmp[1] = ':';
      exec_tmp[2] = '\0';
      strcat(exec_tmp,execname);
#endif
      Trace(("/ -\nexec_tmp\t%s\nexecname\t%s\nexecpath\t%s\n",
             exec_tmp,execname,execpath));
      if (! fileExistsMulti(exec_tmp2,NULL,exec_tmp,exts,NUM_EXTENSIONS) )
      {
          j = 0;
          for (i = 0; i < NUM_EXTENSIONS; i++)
              j += sprintf(buf + j," [%d]: %s\n",i+1,exts[i]);
          error("Couldn't locate %s\nI tried appending the following "
                "extensions: \n%s",exec_tmp,buf);
      }
      Trace((exec_tmp2));
   }
   /* 
    * STARTS WITH x:\ or x:/
    * execpath NOT used
    */
    else if ((strlen(execname) > 3) && /* avoid boundary errors */
       (execname[1] == ':') &&
       ((execname[2] == '\\') || (execname[2] == '/')))
   {
      strcpy(exec_tmp,execname);
      Trace(("x: -\nexec_tmp\t%s\nexecname\t%s\nexecpath\t%s\n",
             exec_tmp,execname,execpath));
      if (! fileExistsMulti(exec_tmp2,NULL,exec_tmp,exts,NUM_EXTENSIONS) )
      {
          j = 0;
          for (i = 0; i < NUM_EXTENSIONS; i++)
              j += sprintf(buf + j," [%d]: %s\n",i+1,exts[i]);
          error("Couldn't locate %s\nI tried appending the following "
                "extensions: \n%s",exec_tmp,buf);
      }
      Trace((exec_tmp2));
   }
   /* 
    * STARTS WITH ./ or .\
    */
   else if ((execname[0] == '.') &&
            ((execname[1] == '\\') || (execname[1] == '/')))
   {
      if (((char*) getcwd(exec_tmp,MAX_PATH))==NULL)
         error("can't find current working directory");
      if (! fileExistsMulti(exec_tmp2,exec_tmp,&(execname[2]),
                            exts,NUM_EXTENSIONS) )
          if (! fileExistsMulti(exec_tmp2,execpath,&(execname[2]),
                                exts,NUM_EXTENSIONS) )
          {
              j = 0;
              for (i = 0; i < NUM_EXTENSIONS; i++)
                  j += sprintf(buf + j," [%d]: %s\n",i+1,exts[i]);
              error("Couldn't locate %s\n"
                    "I looked in the following directories:\n [1]: %s\n [2]: %s\n"
                    "I also tried appending the following "
                    "extensions: \n%s",execname,exec_tmp,execpath,buf);
          }
      Trace((exec_tmp2));
   }
   /*
    * OTHERWISE, SEARCH PATH (prepend '.' and run.exe's directory)
    * can't use fileExistsMulti because we want to search entire path
    * for exts[0], then for exts[1], etc.
    */
   else
   {
      orig_pathlist = getenv("PATH");
      if ((pathlist = malloc (strlen(orig_pathlist)
                              + strlen(".")
                              + strlen(execpath)+ 3)) == NULL)
         error("internal error - out of memory");
      strcpy(pathlist,".");
      strcat(pathlist,SEP_CHARS);
      strcat(pathlist,execpath);
      strcat(pathlist,SEP_CHARS);
      strcat(pathlist,orig_pathlist);

      Trace((pathlist));
      for (i = 0; i < NUM_EXTENSIONS; i++)
      {
          strcpy(exec_tmp,execname);
          strcat(exec_tmp,exts[i]);
          pfopen(exec_tmp2,exec_tmp,pathlist);
          if (fileExists(NULL,NULL,exec_tmp2))
              break;
          exec_tmp2[0] = '\0';
      }
      Trace(("exec_tmp\t%s\npathlist\t%s\n",exec_tmp2,pathlist));

      free(pathlist);
      if (exec_tmp2[0] == '\0')
      {
          j = 0;
          for (i = 0; i < NUM_EXTENSIONS; i++)
              j += sprintf(buf + j," [%d]: %s\n",i+1,exts[i]);
          error("Couldn't find %s anywhere.\n"
                "I even looked in the PATH \n"
                "I also tried appending the following "
                "extensions: \n%s",execname,buf);
      }
   }
/*
 * At this point, we know that exec_tmp2 contains a filename
 * and we know that exec_tmp2 exists.
 */
#if defined(__CYGWIN__)
   {
      struct stat stbuf;
      char sym_link_name[MAX_PATH+1];
      char real_name[MAX_PATH+1];
      char dummy[MAX_PATH+1];

      strcpy(exec_tmp,exec_tmp2);

      CYGWIN_CONV_TO_POSIX_PATH((exec_tmp,sym_link_name));
      Trace((sym_link_name));

      if (lstat(sym_link_name, &stbuf) == 0)
      {
         if ((stbuf.st_mode & S_IFLNK) == S_IFLNK)
         {
	    if (readlink(sym_link_name, real_name, sizeof(real_name)) == -1)
               error("problem reading symbolic link for %s",exec_tmp);
            else
            {
                /* if realname starts with '/' it's a rootpath */
                if (real_name[0] == '/')
                    strcpy(exec_tmp2,real_name);
                else /* otherwise, it's relative to the symlink's location */
                {
                   CYGWIN_SPLIT_PATH((sym_link_name,exec_tmp2,dummy));
                   if (!endsWith(exec_tmp2,PATH_SEP_CHAR_STR))
                      strcat(exec_tmp2,PATH_SEP_CHAR_STR);
                   strcat(exec_tmp2,real_name);
                }
            }
         }
         else /* NOT a symlink */
            strcpy(exec_tmp2, sym_link_name);
      }
      else
         error("can't locate executable - %s",sym_link_name);
   }
   CYGWIN_CONV_TO_FULL_WIN32_PATH((exec_tmp2,exec));
#else
   strcpy (exec, exec_tmp2);
#endif
}
int endsWith(const char* s1, const char* s2)
{
    int len1;
    int len2;
    int retval = FALSE;
    len1 = strlen(s1);
    len2 = strlen(s2);
    if (len1 - len2 >= 0)
        if (stricmp(&(s1[len1-len2]),s2) == 0)
            retval = TRUE;
    return retval;
}void strip_exe(char* s)
{
   if ((strlen(s) > 4) && /* long enough to have .exe extension */
       /* second part not evaluated (short circuit) if exec_arg too short */
       (stricmp(&(s[strlen(s)-4]),".exe") == 0))
      s[strlen(s)-4] = '\0';
}
void error(char* fmt, ...)
{
   char buf[4096];
   int j;
   va_list args;
   va_start(args, fmt);
   j =   sprintf(buf,    "Error: ");
   j += vsprintf(buf + j,fmt,args);
   j +=  sprintf(buf + j,"\n");
   va_end(args);
   MessageBox(NULL, buf, "Run.exe", MB_ICONSTOP);
   exit(1);
}
void message(char* fmt, ...)
{
   char buf[10000];
   int j;
   va_list args;
   va_start(args, fmt);
   j = vsprintf(buf,fmt,args);
   j +=  sprintf(buf + j,"\n");
   va_end(args);
   MessageBox(NULL, buf, "Run.exe Message", MB_ICONSTOP);
}
void Trace_(char* fmt, ...)
{
   char buf[10000];
   int j;
   va_list args;
   va_start(args, fmt);
   j = vsprintf(buf,fmt,args);
   j +=  sprintf(buf + j,"\n");
   va_end(args);
   MessageBox(NULL, buf, "Run.exe DEBUG", MB_ICONSTOP);
}
/*
 * Uses system info to determine the path used to invoke run
 * Also attempts to deduce the target execname if "compact_invocation"
 * method was used.
 *
 * returns TRUE if compact_invocation method was used
 *   (and target execname was deduced successfully)
 * otherwise returns FALSE, and execname == run or run.exe
 */
int get_exec_name_and_path(char* execname, char* execpath)
{
   char modname[MAX_PATH];
   char* tmp_execname;
   char* p;
   int retval = FALSE;

   if (!GetModuleFileName (NULL, modname, MAX_PATH))
      error("internal error - can't find my own name");
   if ((p = strrchr (modname, '\\')) == NULL)
      error("internal error - my own name has no path\n%s",modname);
   tmp_execname = p + 1;
   p[0] = '\0';
   /* if invoked by a name like "runxemacs" then strip off
    * the "run" and let "xemacs" be execname.
    * To check for this, make that:
    *   1) first three chars are "run"
    *   2) but the string doesn't end there, or start ".exe"
    * Also, set "compact_invocation" TRUE
    */
   if ( ((tmp_execname[0] == 'r') || (tmp_execname[0] == 'R')) &&
        ((tmp_execname[1] == 'u') || (tmp_execname[1] == 'U')) &&
        ((tmp_execname[2] == 'n') || (tmp_execname[2] == 'N')) &&
        ((tmp_execname[3] != '.') && (tmp_execname[3] != '\0')) )
   {
      tmp_execname += 3;
      retval = TRUE;
   }
   else
      tmp_execname = NULL;

   if (tmp_execname == NULL)
      strcpy(execname,"");
   else
      strcpy(execname,tmp_execname);
#if defined(__CYGWIN__)
   CYGWIN_CONV_TO_POSIX_PATH((modname,execpath));
#else
   strcpy(execpath,modname);
#endif
   return retval;
}
/*
 * works like strtok, but:
 * double quotes (") suspends tokenizing until closing " reached
 * CYGWIN ONLY:
 *   additionally, backslash escapes next character, even if that
 *   next character is a delimiter. Or a double quote.
 *   WARNING: this means that backslash may NOT be a delimiter 
 */
char* my_strtok(char* s, const char* delim, char** lasts)
{
   char *spanp;
   int c, sc;
   char *tok;

   if ((s == NULL) && ((s = *lasts) == NULL))
      return NULL;
   /* Skip leading delimiters */
cont:
   c = *s++;
   for (spanp = (char *)delim; (sc = *spanp++) != 0;) {
      if (c == sc)
         goto cont;
   }
   if (c == 0) {		/* no non-delimiter characters */
      *lasts = NULL;
      return (NULL);
   }
   tok = s - 1;
   /*
    * Scan token (scan for delimiters: s += strcspn(s, delim), sort of).
    * Note that delim must have one NUL; we stop if we see that, too.
    * If we see a double quote, continue until next double quote, then
    *   start scanning for delimiters again.
    * CYGWIN ONLY: if we see a backslash, just copy next character -
    *   don't consider it as a delimiter even if it is in delim string.
    */
   for (;;) {
      /* if this c is ", then scan until we find next " */
      if (c == '\"')
         while ((c = *s++) != '\"')
            if (c == 0) /* oops, forgot to close the ", clean up & return */
            {
               s = NULL;
               *lasts = s;
               return (tok);
            }
#if defined(__CYGWIN__)
      if (c == '\\')
      {
         c = *s++; /* skip the backslash */
         if (c == 0) /* if escaped character is end-of-string, clean up & return */
         {
            s = NULL;
            *lasts = s;
            return (tok);
         }
         c = *s++; /* otherwise, skip the escaped character */
      }
#endif
      spanp = (char *)delim;
      do {
         if ((sc = *spanp++) == c) {
            if (c == 0)
               s = NULL;
            else
               s[-1] = 0;
            *lasts = s;
            return (tok);
         }
      } while (sc != 0);
      c = *s++;
   }
   /* NOTREACHED */
}

#ifdef __CYGWIN__
void addPathCygwin(const char *str)
{
    char wstr[MAX_PATH];
    DWORD envsize;
    size_t slen;
    char *buffer;

    CYGWIN_CONV_TO_FULL_WIN32_PATH((str,wstr));
    slen = strlen(wstr);

    envsize = GetEnvironmentVariable("PATH", NULL, 0);
    if (envsize == 0)
    {
        DWORD lasterror = GetLastError();
        if (lasterror == ERROR_ENVVAR_NOT_FOUND)
        {
            if (!SetEnvironmentVariable("PATH", wstr))
                error("SetEnvironmentVariable failed: 0x%x", GetLastError());
        }
        error("GetEnvironmentVariable failed: 0x%x", lasterror);
    }
    buffer = malloc(envsize + slen + 2);
    if (buffer == NULL)
        error("internal error - out of memory");

    if (!GetEnvironmentVariable("PATH", buffer, envsize))
        error("GetEnvironmentVariable failed: 0x%x", GetLastError());

    strcat(buffer, ";");
    strcat(buffer, wstr);

    if (!SetEnvironmentVariable("PATH", buffer))
        error("SetEnvironmentVariable failed: 0x%x", GetLastError());
    free(buffer);
}

#endif
void addPath(const char *str)
{
    const char *path = getenv("PATH");
    if (path == NULL)
    {
        run_setenv("PATH", str);
    } else
    {
        char *buffer = malloc(strlen(path) + strlen(str) + 2);
        strcpy(buffer, path);
        strcat(buffer, ":");
        strcat(buffer, str);
        run_setenv("PATH", buffer);
        free(buffer);
    }
#ifdef __CYGWIN__
    addPathCygwin(str);
#endif
}


int parse_cmdline_to_arg_array(char* argv[MAX_ARGS], char* cmdline)
{
   char seps[] = " \t\n";
   char* token;
   int argc = 0;
   char* lasts;

   token = my_strtok(cmdline, seps, &lasts);
   while ((token != NULL) && (argc < MAX_ARGS))
   {
       if (strcmp(token,"-p") == 0)
       {
           token = my_strtok(NULL,seps,&lasts);
           if (token == NULL)
           {
               error("-p expects parameter");
           }

           if (*token == '"')
           {
               int l;
               token++;
               l = strlen(token);
               if (l > 0 && token[l-1] == '"')
                   token[l-1] = 0;
           }
           addPath(token);
       } else
       {
           if ((argv[argc] = malloc(strlen(token)+1)) == NULL)
           {
               error("internal error - out of memory");
           }
           strcpy(argv[argc++],token);
       }
       token = my_strtok(NULL,seps,&lasts);
   }
   if (argc >= MAX_ARGS)
      error("too many arguments on commandline\n%s",cmdline);
   return argc;
}
/* Taken from pfopen.c by David Engel (5-Jul-97).
 * Original comments appear below. Superseded by next comment block.
 *
 *  Written and released to the public domain by David Engel.
 *
 *  This function attempts to open a file which may be in any of
 *  several directories.  It is particularly useful for opening
 *  configuration files.  For example, PROG.EXE can easily open
 *  PROG.CFG (which is kept in the same directory) by executing:
 *
 *      cfg_file = pfopen("PROG.CFG", "r", getenv("PATH"));
 *
 *  NULL is returned if the file can't be opened.
 */

/*
 * This function attempts to locate a file which may be in any of
 * several directories. Unlike the original pfopen, it does not
 * return a FILE pointer to the opened file, but rather returns
 * the fully-qualified filename of the first match found. Returns
 * empty string if not found.
 */
char *pfopen(char *retval, const char *name, const char *dirs)
{
    char *ptr;
    char *tdirs;
    char returnval[MAX_PATH + FILENAME_MAX + 100];
    char *recursive_name;
    int foundit = FALSE;

    returnval[0] = '\0';

    if (dirs == NULL || dirs[0] == '\0')
        return NULL;

    if ((tdirs = malloc(strlen(dirs)+1)) == NULL)
        return NULL;

    strcpy(tdirs, dirs);

    for (ptr = strtok(tdirs, SEP_CHARS); (foundit == FALSE) && ptr != NULL;
         ptr = strtok(NULL, SEP_CHARS))
    {
       foundit = fileExists(returnval,ptr,name);
    }

    free(tdirs);
    if (!foundit)
        retval[0] = '\0';
    else
        strcpy(retval,returnval);
    return retval;
}
int fileExistsMulti(char* fullname, const char* path,
                    const char* name_noext, const char* exts[],
                    const int extcnt)
{
    char tryName[MAX_PATH + FILENAME_MAX];
    int i = 0;
    int retval = FALSE;
    fullname[0] = '\0';
    for (i = 0; i < extcnt; i++)
    {
        strcpy(tryName,name_noext);
        strcat(tryName,exts[i]);
        if (fileExists(fullname, path, tryName) == TRUE)
        {
            retval = TRUE;
            break;
        }
        fullname[0] = '\0';
    }
    return retval;
}
int fileExists(char* fullname, const char* path, const char* name)
{
   int retval = FALSE;
   FILE* file;
   size_t len;
   char work[FILENAME_MAX];
   char work2[MAX_PATH + FILENAME_MAX + 100];
   size_t wlen = sizeof(work);
   if (path != NULL)
   {
      strncpy(work, path, wlen);
      work[sizeof(work)-1] = 0;
      len = strlen(work);
      wlen = sizeof(work) - len;
      if (len && work[len-1] != '/' && work[len-1] != '\\')
      {
         strncat(work, PATH_SEP_CHAR_STR, wlen);
         work[sizeof(work)-1] = 0;
         wlen = sizeof(work) - strlen(work);
      }
   }
   else
      work[0]='\0';

   strncat(work, name, wlen);
   work[sizeof(work)-1] = 0;
   wlen = sizeof(work) - strlen(work);
#if defined(__CYGWIN__)
   CYGWIN_CONV_TO_POSIX_PATH((work, work2));
#else
   strcpy(work2,work);
#endif

#ifdef DEBUGALL
   Trace(("looking for...\t%s\n",work2));
#endif

   file = fopen(work2, "rb");
   if (file != NULL)
   {
      if (fullname != NULL)
         strcpy(fullname,work2);
      retval = TRUE;
      fclose(file);
   }
   return retval;
}

void
run_setenv (const char *name, const char *value)
{
  {
#ifdef HAVE_SETENV
    setenv (name, str, 1);
#else
    int len = strlen (name) + 1 + strlen (value) + 1;
    char *str = (char *) malloc (len);
    sprintf (str, "%s=%s", name, value);
    if (putenv (str) != EXIT_SUCCESS)
      {
        free (str);
      }
#endif
  }
}

